/*
 * --------------------------------------------------------------------------------------------------------------------
 * Detection de trains par lecture d'etiquettes RFID RC-522 Ultralight 
 * --------------------------------------------------------------------------------------------------------------------
 * RFID Tag or Card using a MFRC522 based RFID
 * Reader on the Arduino SPI interface.
 * 
 * 
 * Copyright Locoduino et Dominique Bultez 2020.
 * 
 * Typical pin layout used:
 * -----------------------------------------------------------------------------------------
 *             MFRC522      Arduino       Arduino   Arduino    Arduino          Arduino
 *             Reader/PCD   Uno/101       Mega      Nano v3    Leonardo/Micro   Pro Micro
 * Signal      Pin          Pin           Pin       Pin        Pin              Pin
 * -----------------------------------------------------------------------------------------
 * RST/Reset   RST          9             5         D9         RESET/ICSP-5     RST
 * SPI SS      SDA(SS)      10            53        D10        10               10
 * SPI MOSI    MOSI         11 / ICSP-4   51        D11        ICSP-4           16
 * SPI MISO    MISO         12 / ICSP-1   50        D12        ICSP-1           14
 * SPI SCK     SCK          13 / ICSP-3   52        D13        ICSP-3           15
 * 
 * Pour 2 lecteurs SPI :
 * SPI SS1 = 8
 * SPI SS2 = 7
 * 
 * exemple de lecture de carte ou etiquette
 * Reader 0: Card UID: B6 E4 9C F7                - >  PICC type: MIFARE 1KB
 * Reader 1: Card UID: 04 ED 7E C2 99 58 80       - >  PICC type: MIFARE Ultralight or Ultralight C
 * 
 * Le dernier NUID lu est conservé en mémoire pour interrogation via Can
 * Le lecture des données est dans le secteur #1 (blocks #4 à #7).
 * Format :
 * NN BB63259xxxxxx
 * NN numéro de loco ou train (l'étiquette n'est pas forcément sous la loco),
 * puis un espace séparateur,
 * puis le nom du train ou loco,

 * messages Can emis Id (court) = 0x1A (detecteur EST)
 * octet 0 : bit 7 = 1, bit 6 = 0,            -> numero de train = bits 0 à 5 (0 à 63) zone 25 (interieur)
 * octet 0 : but 7 = 0, bit 6 = 1             -> numero de train = bits 0 à 5 (0 à 63) zone 16 (exterieur)
 * messages Can emis Id (court) = 0x1B (detecteur OUEST)
 * octet 0 : bit 7 = 1, bit 6 = 0,            -> numero de train = bits 0 à 5 (0 à 63) zone 30 (interieur)
 * octet 0 : but 7 = 0, bit 6 = 1             -> numero de train = bits 0 à 5 (0 à 63) zone 11 (exterieur)
 * et
 * La 1ère fois le NUID est affecté au 1er index vide dans le tableau
 * octets 1 à 7 = nom de loco ou train (limité à 7 caractères)

 * Configuration
 * messages Can reçus Id (long) = Id 0x1FFFFF60 (hex) + ID court (1A ou 1B)
 * 
 * 
 * configuration EEPROM (1024 bytes max)
 * start = 0x00
 * byte version 011 
 * byte 0x1A Id EST ou 0x1B id OUEST
 */
#include <SPI.h>
#include <mcp_can.h>
#include <MFRC522.h>
#include <ACAN2515Settings.h> /* Pour les reglages  */
#include <ACAN2515.h>         /* Pour le controleur */
#include <CANMessage.h>       /* Pour les messages  */
#include <EEPROM.h>         

#define SS_1_PIN 8 
#define SS_2_PIN 7 
#define RST_PIN 9
#define NR_OF_READERS   2
byte ssPins[] = {SS_1_PIN, SS_2_PIN};

MFRC522 mfrc522[NR_OF_READERS];   // Create MFRC522 instances.
MFRC522::StatusCode status;       // variable pour le statut des opérations 

uint8_t pageAddr = 0x04;      //Les données sont ecrites dans 4 pages : 4,5,6 et 7.
                              //Pages 0 to 3 sont réservées donc inutilisées.       
const byte versionNb = 13;    // 0.1.3 du 22/06/20 Tout changement de version provoque une Raz de l"EEPROM 
const byte OUEST = 0x1A;      // 26
const byte EST = 0x1B;        // 27
const byte EstouOuest = EST;  // position des lecteurs

static const byte MCP2515_CS  = 10 ; // CS input of MCP2515
static const byte MCP2515_INT = 3 ; // INT output of MCP2515

ACAN2515 controleurCAN(MCP2515_CS, SPI, MCP2515_INT);
static const uint32_t FREQUENCE_DU_QUARTZ = 16ul * 1000ul * 1000ul; //16Mhz
static const uint32_t FREQUENCE_DU_BUS_CAN = 500ul * 1000ul;        //500kHz
CANMessage messageCANEmission;
CANMessage messageCANReception;

void messageCmd (const CANMessage & inMessage) {
  uint32_t mId = inMessage.id;
  uint8_t cmdType = inMessage.data[0];
  if ((mId & 0x1F) == EstouOuest) { // test 1A (EST) ou 1B (OUEST)
    
  }
}

/* 
 * Définition des masques et filtres.
 */
const ACAN2515Mask masque = extended2515Mask (0x1FFFFFFF) ; // For filter #0 and #1
  const ACAN2515AcceptanceFilter filtres [] = {
    {extended2515Filter (0x1FFFFF7A), messageCmd},
    {extended2515Filter (0x1FFFFF7B), messageCmd},
  } ;


// EEPROM image
const char *EEPROM_ID = { "RFID" };
byte detectId = EstouOuest;  
// EST =  0x1B ou 27;  Interieur z25, Exterieur z16
// OUEST = 0x1A ou 26; Interieur z30, Exterieur z11
byte isEstOuest;  //1->1A Est ou 0->0x1B Ouest


void EEPROM_sauvegarde() {
  int addr = 0;
  for ( int i = 0; i<4; i++) {
  EEPROM.update(addr++, EEPROM_ID[i]);
  }
  EEPROM.update(addr++, versionNb);
  EEPROM.update(addr++, detectId);
  EEPROM.update(addr++, isEstOuest);
}

bool EEPROM_chargement() {
  int addr = 0;
  char buf[4];
  byte vers;
  for (int i = 0; i < 4; i++) {
    buf[i] = EEPROM.read(addr++);
    if (buf[i] != EEPROM_ID[i]) return false;
  }
  vers = EEPROM.read(addr++);
  Serial.print("version EEPROM ");Serial.println(vers);
  if (vers != versionNb) return false;
  detectId = EEPROM.read(addr++);
  Serial.print("detectId EEPROM ");Serial.println(detectId);
  if (detectId == 27) isEstOuest = 1; else isEstOuest = 0;
  isEstOuest = EEPROM.read(addr++);
  Serial.print("isEstOuest EEPROM ");Serial.println(isEstOuest? "EST": "OUEST");
  return true;
}

void setup() { 
  Serial.begin(115200);
  while (!Serial);    // Do nothing if no serial port is opened (added for Arduinos based on ATMEGA32U4)
  
  SPI.begin(); // Init SPI bus
  
  for (uint8_t reader = 0; reader < NR_OF_READERS; reader++) {
    mfrc522[reader].PCD_Init(ssPins[reader], RST_PIN); // Init each MFRC522 card
    Serial.print(F("Reader "));
    Serial.print(reader);
    Serial.print(F(": "));
    mfrc522[reader].PCD_DumpVersionToSerial();
  }
  Serial.println("RC522 init OK");
  
  ACAN2515Settings configuration(FREQUENCE_DU_QUARTZ, FREQUENCE_DU_BUS_CAN);
  //const uint32_t codeErreur = controleurCAN.begin(configuration, [] { controleurCAN.isr(); }); // ni masque, ni filtre
  const uint16_t codeErreur = controleurCAN.begin(configuration, [] { controleurCAN.isr(); }, masque, filtres, 2 );
  if (codeErreur == 0) {
  Serial.println("controleurCAN configuration OK");
  } else {
    Serial.print("controleurCAN: Probleme de connexion: ");
    Serial.println(codeErreur);
  }
  isEstOuest = (EstouOuest == EST)? true : false;
  Serial.print("versionSW " ); Serial.print(versionNb); Serial.print(" Position " ); Serial.print(EstouOuest, HEX);
  Serial.print(" ->" ); Serial.println(isEstOuest? "EST" : "OUEST");
  Serial.println("chargement EEPROM" );
  
  if (EEPROM_chargement() == false) {

    EEPROM_sauvegarde();
  }
    
  //Serial.println(F("1:dump, 2:Raz, 3:sauve, 4:lire eeprom"));
} // end of setup


/**
 * Routines d'affichage d'un tableau d'octets en Hexa et en Ascii (char)
 */
void dump_byte_array(byte *buffer, byte bufferSize) {
    for (byte i = 0; i < bufferSize; i++) {
        Serial.print(buffer[i] < 0x10 ? " 0" : " ");
        Serial.print(buffer[i], HEX);
    }
}

void dump_char_array(byte *buffer, byte bufferSize) {
    for (byte i = 0; i < bufferSize; i++) {
        Serial.print("  ");
        Serial.print((char)buffer[i]);
    }
}

void loop() {
  
  controleurCAN.dispatchReceivedMessage();
  byte nloco;
  byte loco;

  for (uint8_t reader = 0; reader < NR_OF_READERS; reader++) {
    // Look for new cards
    if (mfrc522[reader].PICC_IsNewCardPresent() && mfrc522[reader].PICC_ReadCardSerial()) {
      Serial.print("Detection ");Serial.print(isEstOuest? "EST " : "OUEST ");
      Serial.print((reader==0)? "Interieur: " : "Exterieur: "); // reader = 0 ou 1 (1 exterieur et 0 interieur)
      dump_byte_array(mfrc522[reader].uid.uidByte, mfrc522[reader].uid.size);
      Serial.println();
    
      Serial.print(F("PICC type: "));
      MFRC522::PICC_Type piccType = mfrc522[reader].PICC_GetType(mfrc522[reader].uid.sak);
      Serial.println(mfrc522[reader].PICC_GetTypeName(piccType));
      // verification de compatibilité
      if (piccType != MFRC522::PICC_TYPE_MIFARE_UL) {
        Serial.println(F("Ce programme ne fonctionne qu'avec des étiquettes MIFARE Ultralight"));
        return;
      }
     
      // lecture numero et nom du train/loco
      // Nous utilisons les pages 4 à 7 inclus (4 octets/page donc 16 octets)
      byte sector         = 1;
      byte blockAddr      = pageAddr;
      byte trailerBlock   = pageAddr + 3;
      byte buffer[18];        // 16 caractères + CRC
      byte size = sizeof(buffer);
    
      // Affichage du contenu du secteur 
      Serial.print(F("Contenu des pages: "));Serial.print(blockAddr);Serial.print(" à ");Serial.println(trailerBlock);
      status = (MFRC522::StatusCode) mfrc522[reader].MIFARE_Read(blockAddr, buffer, &size);
      if (status != MFRC522::STATUS_OK) {
        Serial.print(F("MIFARE_Read() Echec: "));
        Serial.println(mfrc522[reader].GetStatusCodeName(status));
      }
      dump_byte_array(buffer, 16); Serial.println();
      dump_char_array(buffer, 16); Serial.println();
      Serial.println();

      // numero de loco
      nloco = (buffer[0] - '0')*10 + (buffer[1] - '0');
      Serial.print("Loco No ");Serial.print(nloco);
      // nom de loco
      Serial.print(" - Nom : ");
      dump_char_array(&buffer[3], 8);
      Serial.println();
            
      // preparation envoi message CAN
      if (isEstOuest) messageCANEmission.id = OUEST; else messageCANEmission.id = EST;
      messageCANEmission.ext = false ;
      messageCANEmission.len = 8 ;
      loco = nloco;
      if (reader == 1) {loco = nloco + 0x80;}
      messageCANEmission.data[0] = loco;
      for (byte i = 1; i < 8; i++) {
        messageCANEmission.data[i] =buffer[i+2];
      }

      // envoi du message Can
      const bool ok = controleurCAN.tryToSend(messageCANEmission);
      delay(50);
      if (ok) {
        Serial.print("Message train ");Serial.print(nloco);Serial.println(" transmis !");
      } else {
        Serial.println("Pb de transmission Can !");
      }
      Serial.println();

           
      // Halt PICC
      mfrc522[reader].PICC_HaltA();
      // Stop encryption on PCD
      mfrc522[reader].PCD_StopCrypto1();
    } //if (mfrc522[reader].PICC_IsNewC...
  } //for(uint8_t reader)

  // réception des messages CAN : voir messageCmd
  if (controleurCAN.receive(messageCANReception)) {
    Serial.print("->recu : ");
    Serial.print(messageCANReception.len);
    Serial.print(", data : ");
    Serial.print(messageCANReception.data[0], HEX);
    Serial.print(", ");
    Serial.print(messageCANReception.data[1], HEX);
    Serial.print(", ");
    Serial.println(messageCANReception.data[2], HEX);
    messageCmd (messageCANReception);
  }

  if (Serial.available()) { // commandes console
    char code = (char)Serial.read();
    if (code == '1') {
      
    }
    if (code == '2') {
      
    }
    if (code == '3') {

    }
    if (code == '4') {

    }
  Serial.println(F("1:dump, 2:Raz, 3:sauve, 4:lire eeprom"));
  } // end Serial console

} // end of loop
